package cz.upol.jj2.lab10.reflection;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Map;
import java.util.TreeMap;
import java.util.stream.Collectors;

import cz.upol.jj2.lab10.reflection.examples.Rectangle;
import cz.upol.jj2.lab10.reflection.examples.Useless;

/**
 * prace s anotacemi (standardnimi i vlastnimi)
 * s lehkymi upravami prevzato od dr. Krajci
 */
public class AnnotationReader {
	
	
	public static void main(String[] args) throws NoSuchMethodException, SecurityException, IllegalAccessException, IllegalArgumentException, InvocationTargetException {
		
		Class<?> clazz = Useless.class;
		Annotation[] annots = clazz.getAnnotations();
		
		for (Annotation a : annots) {
			if (a instanceof Deprecated)
				System.out.println("class is deprecated");
		}
		
		Method m = clazz.getMethod("nothingToDo");
		Deprecated deprAnn = m.getAnnotation(Deprecated.class);
		System.out.println("Deprecated since: " + deprAnn.since());

		//ale override anotaci takto nenajdeme - za behu jiz neni dostupna
		m = clazz.getMethod("toString");
		Override overrideAnn = m.getAnnotation(Override.class);
		System.out.println(overrideAnn);

		//ukazka inspect
		Rectangle r = new Rectangle(10, 20, 30, 40);
		System.out.println(r.toString());
		System.out.println(inspectObject(r));
		//... ale neco se stalo s r?
		System.out.println(r.toString());
		//cim to?
	}
	
	
	/**
	 * metoda vrati retezec obsahujici hodnoty atributu a metod,
	 * ktere byly v objektu <tt>obj</tt> oznaceny anotaci {@link Inspect}
	 */
	public static String inspectObject(Object obj) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
		Class<?> clazz = obj.getClass();
		Map<String, String> values = new TreeMap<>();
		
		for (Field field : clazz.getDeclaredFields()) {
			field.setAccessible(true);
			Inspect ann = field.getAnnotation(Inspect.class);
			
			if (ann != null) {
				// atribut ma anotaci, a proto jej precteme
				Object value = field.get(obj);
				String name = ann.as().isEmpty() ? field.getName() : ann.as();
				values.put(name, value.toString());
			}
		}
		
		for (Method method : clazz.getDeclaredMethods()) {
			method.setAccessible(true);
			Inspect ann = method.getAnnotation(Inspect.class);
			
			if (ann != null) {
				// metoda ma anotaci, a proto ji zavolame
				Object value = method.invoke(obj);
				String name = ann.as().isEmpty() ? method.getName() : ann.as();
				values.put(name, value.toString());
			}
		}
		
		String valueStr = values.entrySet().stream()
			.map(e -> e.getKey() + " = " + e.getValue())
			.collect(Collectors.joining("," , "[", "]"));
		
		return clazz.getName() + " " + valueStr;
		
	}

}
