package cz.upol.jj2.lab10.reflection;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.stream.Collectors;

import cz.upol.jj2.lab10.reflection.examples.Animal;
import cz.upol.jj2.lab10.reflection.examples.Employee;

/**
 * Obecna prace s reflexi: ziskavani atributu a metod, volani metod, ...
 * Lehce upravena verze prikladu k textu dr. Krajci
 */
public class ReflectionExample {
	
	public static void main(String[] args) throws Exception {
		Class<Employee> clazz = Employee.class;
		showClassInfo(clazz);

		// prace s tridou Employee
		Object obj = createInstance(clazz);
		System.out.println(obj);
		setAge(obj);
		System.out.println(obj);
 

		Class<Animal> clazzAnimal = Animal.class;
		showClassInfo(clazzAnimal);

		// prace s tridou Animal
		Object obj2 = createInstance(clazzAnimal);
		System.out.println(obj2);
		setAge(obj2);
		System.out.println(obj2);

		//tj. pouzivame stejny kod (createInstance, setAge) pro tridy,
		//ktere z poheldu tridni hierarchie nemaji nic spolecneho (krome object)

		//ale pokud tam dame objekt spatneho typu, tak mame problem.
		//Object obj3 = new Object();
		//setAge(obj3);

	}

	/**
	 * Zobrazi informace o dane tride
	 */
	private static void showClassInfo(Class<?> clazz) {
		System.out.println("Class: " + clazz.getName());
		// vypise seznam vsech deklarovanych atributu a jejich vlastnosti
		System.out.println("\tAttributes:");
	
		Field[] fields = clazz.getDeclaredFields(); //zkuste i clazz.getFields();
		for (Field field : fields) {
			System.out.println(
					String.format("\t\t%s: %s (%s)",
							field.getName(),
							field.getType(),
							Modifier.toString(field.getModifiers())));
		}
		
		// vypise seznam vsech deklarovanych metod
		System.out.println("\tMethods:");

		Method[] methods = clazz.getMethods(); // zkuste i clazz.getMethods();
		for (Method method : methods) {
			System.out.println(
					String.format("\t\t%s %s %s(%s)",
							Modifier.toString(method.getModifiers()),
							method.getReturnType().getName(),
							method.getName(),
							Arrays.stream(method.getParameterTypes())
												.map(l -> l.getName())
												.collect(Collectors.joining(","))));
		}
	}
	
	/**
	 * Vytvori objekt zadane tridy. Predpoklada se, 
	 * ze objekt ma konstruktor bez parametru a metodu setName(String).
	 */
	private static Object createInstance(Class<?> clazz) throws Exception {
		
		// ziska konstruktor (lze uvest i typy parametru)
		Constructor<?> constructor = clazz.getConstructor();
		
		// vytvori novy objekt
		Object obj = constructor.newInstance();

		// ziskani a zavolani metody
		Method setNameMethod = clazz.getMethod("setName", String.class);
		setNameMethod.invoke(obj, "Sova Antonin");
		return obj;
	}
	
	/**
	 * Modifikuje objekt, predpoklada se, ze objekt ma metodu setAge(int). 
	 */
	private static void setAge(Object obj) throws SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		// ziskani a zavolani metody
		Method m = obj.getClass().getMethod("setAge", int.class);
		m.setAccessible(true);
		m.invoke(obj, 50);
		//zkuste to zavolat s objektem, ktery danou metodu nema...
	}

}
